package handlers_test

import (
	"net/http"
	"testing"
	"time"

	"github.com/rs/zerolog"
	"github.com/stretchr/testify/require"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/testutils"
	"orus.io/orus-io/rednerd/testutils/apitester"
)

func AddUserActivity(
	t *testing.T,
	log zerolog.Logger,
	db *database.TestDB,
	username string,
	qtyCurrentMonth, qtyLastMonth int,
) {
	t.Helper()

	renderList := make([]*models.RenderLog, 0, qtyCurrentMonth+qtyLastMonth)

	for range qtyCurrentMonth {
		rd := models.NewRenderLog(time.Now(), "text", "mjml", username)
		rd.SetDuration(time.Second)
		renderList = append(renderList, rd)
	}

	for range qtyLastMonth {
		rd := models.NewRenderLog(time.Now().AddDate(0, -1, 0), "text", "mjml", username)
		rd.SetDuration(time.Second)
		renderList = append(renderList, rd)
	}

	for _, rlog := range renderList {
		_, err := database.Exec(db, database.SQLInsert(rlog), &log)
		require.NoError(t, err)
	}
}

func TestUserGet(t *testing.T) {
	log := testutils.GetLogger(t)
	ctx := t.Context()
	ctx = log.WithContext(ctx)

	tester := apitester.NewAPITester(ctx, t, nil)
	defer tester.Close()

	// login as admin and create a user
	tester.Login("admin", "admin")
	tester.CreateUser("johndoe", "johndoe")

	// adding some dummy activity for the user JohnDoe
	qtyCurrentMonth := 4
	qtyLastMonth := 2
	AddUserActivity(t, log, tester.DB, "johndoe",
		qtyCurrentMonth, qtyLastMonth)

	tester.Logout()

	t.Run("unauthorized", func(t *testing.T) {
		// try and get the list as johndoe_1
		tester.Login("johndoe", "johndoe")
		defer tester.Logout()
		_ = tester.APITest("user_get").
			Get("/api/v1/user/johndoe").
			Expect(t).
			Status(http.StatusUnauthorized).
			End()
	})
	t.Run("basic", func(t *testing.T) {
		tester.Login("admin", "admin")
		defer tester.Logout()
		r := tester.APITest("user_get").
			Get("/api/v1/user/johndoe").
			Expect(t).
			Status(http.StatusOK).
			End()

		var user models.Userinfo

		r.JSON(&user)
		stats := user.Activity.RenderStats
		expectedTotal := qtyCurrentMonth + qtyLastMonth
		require.Equal(t, int64(expectedTotal), stats.Total)
		require.Equal(t, int64(qtyCurrentMonth), stats.CurrentMonth)
		require.Equal(t, int64(qtyLastMonth), stats.LastMonth)
	})
}
