package tests

import (
	"bytes"
	"net/http"
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/testutils/apitester"
)

func TestAuthentication(t *testing.T) {
	ctx := t.Context()

	tester := apitester.NewAPITester(ctx, t, nil)
	defer tester.Close()

	t.Run("WebToken", func(t *testing.T) {
		defer tester.SetT(t)()
		defer tester.Logout()

		tester.Login("admin", "admin")
		userinfo := tester.GetUserInfo()
		assert.Equal(t, "admin", userinfo.Username)
		assert.True(t, userinfo.IsAdmin)
	})

	t.Run("Cookie", func(t *testing.T) {
		defer tester.SetT(t)()
		defer tester.SetHeader("Cookie", "")

		req, _ := http.NewRequestWithContext(
			ctx,
			http.MethodPost,
			"/api/v1/auth/authenticate",
			bytes.NewBufferString(
				`{"username":"admin","password":"admin","set-cookie":true}`,
			),
		)
		req.Header.Set("Content-Type", "application/json")

		res := tester.HandleRequest(req)

		result := res.Result()
		defer result.Body.Close()

		t.Log(res.Header().Get("Set-Cookie"))

		for _, cookie := range result.Cookies() {
			require.Equal(t, "rednerd-auth", cookie.Name)
			tester.SetHeader("Cookie", cookie.String())
		}

		userinfo := tester.GetUserInfo()
		assert.Equal(t, "admin", userinfo.Username)
		assert.True(t, userinfo.IsAdmin)

		req, _ = http.NewRequestWithContext(
			ctx,
			http.MethodPost,
			"/api/v1/auth/logout",
			&bytes.Buffer{},
		)
		req.Header.Set("Content-Type", "application/json")

		res = tester.HandleRequest(req)
		assert.Equal(t, 204, res.Code)
		assert.Equal(t, "", res.Body.String())

		result = res.Result()
		defer result.Body.Close()
		assert.Equal(t, -1, result.Cookies()[0].MaxAge)
		assert.Equal(t, "rednerd-auth", result.Cookies()[0].Name)
	})

	t.Run("APIKey", func(t *testing.T) {
		defer tester.SetT(t)()
		defer tester.Logout()
		defer tester.SetHeader("Rednerd-API-Key", "")

		tester.Login("admin", "admin")
		tester.CreateUser("janedoe", "janedoe")

		var apikey models.ApikeyAndValue

		require.Equal(t, http.StatusCreated,
			tester.MustPostJSON("/api/v1/user/janedoe/apikey", &models.Apikey{
				Name: "key1",
			}, &apikey))
		tester.Logout()

		tester.SetHeader("Rednerd-API-Key", apikey.Value)

		userinfo := tester.GetUserInfo()
		assert.Equal(t, "janedoe", userinfo.Username)
		assert.False(t, userinfo.IsAdmin)
	})
}
