package tests

import (
	"encoding/base64"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/testutils/apitester"
)

func TestRender(t *testing.T) {
	ctx := t.Context()

	tester := apitester.NewAPITester(ctx, t, nil)
	defer tester.Close()

	defer tester.Logout()
	tester.Login("admin", "admin")

	tester.CreateUser("janedoe", "janedoe")
	tester.Login("janedoe", "janedoe")

	t.Run("MJML-Template-to-html", func(t *testing.T) {
		defer tester.SetT(t)()

		var result []*models.Document

		tester.Render(
			&models.RenderRequest{
				Accept: "text/html",
				Template: &models.Template{
					Language:   "mustache",
					Produces:   "text/mjml",
					BodyFormat: "text",
					Body:       "<mjml> <mj-body> <mj-section> <mj-column> <mj-text>hello {{name}}</mj-text> </mj-column> </mj-section> </mj-body> </mjml>", //nolint:lll
				},
				Data: models.Dataset{
					apitester.JSONObj{"name": "Nath"},
				},
			},
			&result,
		)
		require.Equal(t, 1, len(result))

		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello Nath")
		}

		time.Sleep(100 * time.Millisecond)

		var logs []models.RenderLog

		require.NoError(t,
			tester.DB.Select(
				&logs,
				"SELECT * FROM render_logs",
			),
		)

		assert.Equal(t, 1, len(logs))
	})

	t.Run("MJML-Saved-Template-to-html", func(t *testing.T) {
		defer tester.SetT(t)()

		var response models.Template

		tester.CreateTemplate(
			"janedoe",
			&models.Template{
				Name:       "t1",
				Language:   "mustache",
				Produces:   "text/mjml",
				BodyFormat: "text",
				Body:       "<mjml> <mj-body> <mj-section> <mj-column> <mj-text>hello {{name}}</mj-text> </mj-column> </mj-section> </mj-body> </mjml>", //nolint:lll
			},
			&response)

		var result []*models.Document

		tester.Render(
			&models.RenderRequest{
				Accept: "text/html",
				Template: &models.Template{
					Account: "janedoe",
					Name:    "t1",
				},
				Data: models.Dataset{
					apitester.JSONObj{"name": "Nath"},
				},
			},
			&result)
		require.Equal(t, 1, len(result))

		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello Nath")
		}
	})

	t.Run("MJML-with-errors", func(t *testing.T) {
		defer tester.SetT(t)()

		var result []*models.Document

		tester.Render(
			&models.RenderRequest{
				Accept: "text/html",
				Document: &models.Document{
					Type:       "text/mjml",
					BodyFormat: "test",
					Body: `<mjml>
							  <mj-body>
								<mj-section invalid="hi">
								  <mj-column>
									<mj-text invalid=true>hello world</mj-text>
								  </mj-column>
								</mj-section>
							  </mj-body>
							</mjml>`,
				},
			},
			&result,
		)
		require.Equal(t, 1, len(result))

		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello world")
		}

		t.Log(result[0].RenderErrors)
		require.Equal(t, 6, len(result[0].RenderErrors))
		assert.Equal(t, "MJML", result[0].RenderErrors[0].Engine)
		assert.Equal(t, "MJML", result[0].RenderErrors[1].Engine)
	})
	t.Run("MJML-skip-validation", func(t *testing.T) {
		defer tester.SetT(t)()

		var result []*models.Document

		tester.Render(
			&models.RenderRequest{
				Accept: "text/html",
				Metadata: models.Metadata{
					"mjml-validation-level": "skip",
				},
				Document: &models.Document{
					Type:       "text/mjml",
					BodyFormat: "test",
					Body: `<mjml>
							  <mj-body>
								<mj-section invalid="hi">
								  <mj-column>
									<mj-text invalid=true>hello world</mj-text>
								  </mj-column>
								</mj-section>
							  </mj-body>
							</mjml>`,
				},
			},
			&result,
		)
		require.Equal(t, 1, len(result))

		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "hello world")
		}

		require.Equal(t, 0, len(result[0].RenderErrors))
	})

	t.Run("MJML-specific-test", func(t *testing.T) {
		defer tester.SetT(t)()

		var result []*models.Document

		tester.Render(
			&models.RenderRequest{
				Accept: "text/html",
				Document: &models.Document{
					Type:       "text/mjml",
					BodyFormat: "text",
					Body:       mjmlSource,
				},
			},
			&result,
		)
		require.Equal(t, 1, len(result))

		if body, err := base64.StdEncoding.DecodeString(result[0].Body); assert.NoError(t, err) {
			assert.Contains(t, string(body), "<strong>Client</strong> : client")
		}
	})
}

const mjmlSource = `
<mjml>

  <mj-head>

    <mj-attributes>

	<mj-all font-family="Arial" font-size="16px" />
	<mj-text color="#000" line-height="28px" />

	<mj-class name="main-title" font-size="22px" font-weight="bold" color="#969494" />
	<mj-class name="subtitle" font-weight="bold" />
	<mj-class name="primary-color" color="#C32427" />
	<mj-class name="secondary-color" color="#003E8B" />
	<mj-class name="small-italic" font-style="italic" font-size="14px" line-height="18px"/>

	<mj-class name="divider" border-width="1px" border-style="solid" border-color="lightgrey" />
	<mj-class name="border-content" border="1px solid #dedede" padding-top="10px" padding-bottom="10px" />
	<mj-class name="cta-btn" inner-padding="22px 44px" background-color="#003E8B" color="white" font-size="14px" />

	<mj-class name="borded-left" border-left="1px solid #dedede" />
	<mj-class name="borded-left__text" padding="5px 20px" />

	<mj-class name="no-vertical-padding" padding-top="0" padding-bottom="0" />

</mj-attributes>

<mj-style inline="inline">

	a{
		color: #000;
	}
	.link-white {
		color:#fff;
		text-decoration: none;
	}
	.mj-table th, .mj-table td{
		border: 1px solid #ecedee;
		text-align: left;
		padding: 8px 16px;
	}
	.secondary-color{
		color: #003E8B !important;
	}
	.bold{
		font-weight: bold !important;
	}

</mj-style>

<mj-style>

	@media all and (max-width: 480px) { .mobile-center *{ text-align: center !important; }

</mj-style>

  </mj-head>

  <mj-body background-color="#f5f5f5" width="650px">

    <mj-section>

  <mj-column width="30%">
  </mj-column>

  <mj-column width="70%">
    <mj-text font-size="16px" font-weight="bold" align="right" css-class="mobile-center">
      JAL-PRO - ORDRE DE PUBLICATION
    </mj-text>
  </mj-column>

</mj-section>

    <mj-section background-color="#fff">

      <mj-column>

        <mj-wrapper padding="0 20px" background-color="#fff">
          <mj-section>
            <mj-column mj-class="border-content">
              <mj-text align="left">
                <strong>Client</strong> : client<br>
              </mj-text>
            </mj-column>
          </mj-section>
        </mj-wrapper>

      </mj-column>
    </mj-section>
</mj-wrapper>

  </mj-body>

</mjml>

`
