package tests

import (
	"net/http"
	"testing"
	"time"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/strfmt"
	"github.com/rs/zerolog"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/testutils"
	"orus.io/orus-io/rednerd/testutils/apitester"
)

func AddTemplateActivity(
	t *testing.T,
	username string,
	tID uint64,
	qtyCurrentMonth, qtyLastMonth int,
	log zerolog.Logger,
) []*models.RenderLog {
	t.Helper()

	renderList := make([]*models.RenderLog, 0, qtyCurrentMonth+qtyLastMonth)

	for range qtyCurrentMonth {
		rd := models.NewRenderLog(time.Now(), "text", "mjml", username)
		rd.SetDuration(time.Second)
		rd.SetTemplate(tID, 0, "")
		renderList = append(renderList, rd)
	}

	for range qtyLastMonth {
		rd := models.NewRenderLog(time.Now().AddDate(0, -1, 0), "text", "mjml", username)
		rd.SetDuration(time.Second)
		rd.SetTemplate(tID, 0, "")
		renderList = append(renderList, rd)
	}

	return renderList
}

func TestTemplateManagement(t *testing.T) {
	log := testutils.GetLogger(t)
	ctx := t.Context()
	ctx = log.WithContext(ctx)

	tester := apitester.NewAPITester(ctx, t, nil)
	defer tester.Close()

	defer tester.Logout()

	tester.Login("admin", "admin")

	tester.CreateUser("janedoe", "janedoe")
	tester.Login("janedoe", "janedoe")

	db := database.NewSQLHelper(ctx, tester.DB, log)

	qtyCurrentMonth := 2
	qtyLastMonth := 3

	var createdAt strfmt.DateTime

	_ = t.Run("Create", func(t *testing.T) {
		defer tester.SetT(t)()
		var template models.Template
		tester.CreateTemplate(
			"janedoe",
			&models.Template{
				Name:       "t1",
				Language:   "mustache",
				Produces:   "text/html",
				BodyFormat: "text",
				Body:       "<html><body>Hi {{name}}!</body></html>",
			}, &template)

		// checking some values
		assert.Equal(t, "t1", template.Name)
		assert.Equal(t, "janedoe", template.Account)
		assert.NotNil(t, template.CreatedAt)
		createdAt = template.CreatedAt

		// making sure the activity of the template is empty
		// before we add some
		require.NotNil(t, template.Activity)
		stats := template.Activity.RenderStats
		require.NotNil(t, stats)
		require.Equal(t, int64(0), stats.Total)
		require.Equal(t, int64(0), stats.CurrentMonth)
		require.Equal(t, int64(0), stats.LastMonth)

		// we get the template for the DB to get TemplateID
		var tDBList []models.TemplateDB

		require.NoError(t, db.Select(&tDBList,
			squirrel.Select(models.TemplateDBColumns...).
				From(models.TemplateDBTable).
				Where(
					squirrel.Eq{
						models.TemplateDBAccountColumn: template.Account,
					}),
		))
		// checking that we have only one template here
		require.Equal(t, 1, len(tDBList))
		tDB := tDBList[0]

		renderList := AddTemplateActivity(
			t, template.Account, tDB.ID, qtyCurrentMonth, qtyLastMonth, log)
		for _, rlog := range renderList {
			_, err := db.Insert(rlog)
			require.NoError(t, err)
		}
	}) &&
		t.Run("Get_notfound", func(t *testing.T) {
			defer tester.SetT(t)()
			var err models.Error
			require.Equal(t, http.StatusNotFound,
				tester.MustGetJSON("/api/v1/template/janedoe/missing", &err),
			)
		}) &&
		t.Run("Get", func(t *testing.T) {
			defer tester.SetT(t)()
			var template models.Template
			require.Equal(t, http.StatusOK,
				tester.MustGetJSON("/api/v1/template/janedoe/t1", &template),
			)
			assert.Equal(t, "t1", template.Name)
			assert.Equal(t, "janedoe", template.Account)
			assert.Equal(t, createdAt.String(), template.CreatedAt.String())

			// now we check that the activity has been updated
			require.NotNil(t, template.Activity)
			stats := template.Activity.RenderStats
			require.NotNil(t, stats)
			expectedTotal := qtyCurrentMonth + qtyLastMonth
			require.Equal(t, int64(expectedTotal), stats.Total)
			require.Equal(t, int64(qtyCurrentMonth), stats.CurrentMonth)
			require.Equal(t, int64(qtyLastMonth), stats.LastMonth)
		}) &&
		t.Run("List", func(t *testing.T) {
			defer tester.SetT(t)()
			var templateList []models.Template
			require.Equal(t, http.StatusOK,
				tester.MustGetJSON("/api/v1/template/janedoe", &templateList))
			assert.Equal(t, 1, len(templateList))
			assert.Equal(t, "t1", templateList[0].Name)
		}) &&
		t.Run("Update", func(t *testing.T) {
			defer tester.SetT(t)()
			var template models.Template
			tester.UpdateTemplate("janedoe", "t1",
				&models.Template{
					Language:   "mustache",
					Produces:   "text/mjml",
					BodyFormat: "text",
					Body:       "<mjml><mj-text>hello {{name}}</mj-text></mjml>",
				}, &template)
			assert.Equal(t, "t1", template.Name)
			assert.Equal(t, "janedoe", template.Account)
			assert.Equal(t, "text/mjml", template.Produces)
			assert.Equal(t, createdAt.String(), template.CreatedAt.String())
		}) &&
		t.Run("Rename", func(t *testing.T) {
			defer tester.SetT(t)()
			var template models.Template
			tester.UpdateTemplate("janedoe", "t1",
				&models.Template{
					Name:       "t2",
					Language:   "mustache",
					Produces:   "text/html",
					BodyFormat: "text",
					Body:       "<html><body>Hi {{name}}!</body></html>",
				}, &template)
			assert.Equal(t, "t2", template.Name)
			assert.Equal(t, "janedoe", template.Account)
			assert.Equal(t, createdAt.String(), template.CreatedAt.String())

			require.Equal(t, http.StatusOK,
				tester.MustGetJSON("/api/v1/template/janedoe/t2", &template),
			)
			assert.Equal(t, "t2", template.Name)
			assert.Equal(t, "janedoe", template.Account)
		})
}
