package apitester

import (
	"bytes"
	"encoding/json"
	"io"
	"net/http"
	"net/http/httptest"
	"testing"
	"time"

	"github.com/rs/zerolog"
	"github.com/steinfletcher/apitest"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"golang.org/x/net/context"
	"orus.io/orus-io/go-orusapi/database"

	"orus.io/orus-io/rednerd/auth"
	"orus.io/orus-io/rednerd/cmd/rednerd/cmd"
	"orus.io/orus-io/rednerd/engines/wpd"
	"orus.io/orus-io/rednerd/migration"
	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/rendering"
	"orus.io/orus-io/rednerd/restapi"
	"orus.io/orus-io/rednerd/restapi/operations"
	"orus.io/orus-io/rednerd/testutils"
)

// JSONObj is a map.
type JSONObj map[string]interface{}

// APITester provides facilities to unittest the RednerD rest api.
type APITester struct {
	t       *testing.T
	handler http.Handler
	api     *operations.RednerdAPI

	headers map[string]string

	TestLogger *testutils.TestLogger
	Logger     zerolog.Logger
	DB         *database.TestDB
	now        *time.Time

	ctxcancel func()

	Config restapi.APIConfig

	debug bool
}

type ServerOption func(*restapi.Server)

// NewAPITester setup a APITester.
// false positive on thelper that I can't figure out.
// it should allow t *testing.T param to be second here but weirdly does not.
func NewAPITester( //nolint:thelper
	ctx context.Context,
	t *testing.T,
	engineOptions *cmd.EngineOptions,
	serverOptions ...ServerOption,
) *APITester {
	t.Helper()

	testerCtx, cancel := context.WithCancel(ctx)

	log := testutils.GetLogger(t)
	testLog := testutils.NewTestLogger(t)
	testerCtx = log.WithContext(testerCtx)

	var success bool

	db := database.GetTestDB(testerCtx, t, migration.Source)

	defer func() {
		if !success {
			db.Close()
		}
	}()

	tx := db.MustBegin()
	require.NoError(t, cmd.SetAdminPassword(testerCtx, tx, []byte("admin")))
	require.NoError(t, tx.Commit())

	rendering.DefaultEngineRegistry = rendering.NewEngineRegistry()
	rendering.DefaultTemplateEngineRegistry = rendering.NewTemplateEngineRegistry()

	if engineOptions == nil {
		o := cmd.NewEngineOptions()
		engineOptions = &o
		require.NoError(t, engineOptions.SetupEngineOptions(nil))
	}

	wpdOptions, ok := engineOptions.GetOptions("WPD").(*wpd.EngineOptions)
	require.True(t, ok)

	if wpdOptions.WorkerPoolSize == 0 {
		wpdOptions.WorkerPoolSize = 1
	}

	require.NoError(t, engineOptions.SetupEngines(log))

	serveCmd := cmd.NewServerCmd(&cmd.Options{})

	for _, o := range serverOptions {
		o(serveCmd.Server)
	}

	config := restapi.APIConfig{
		Context:                         testerCtx,
		Log:                             log,
		DB:                              db.DB,
		Now:                             time.Now,
		RenderingEngineRegistry:         rendering.DefaultEngineRegistry,
		TemplateRenderingEngineRegistry: rendering.DefaultTemplateEngineRegistry,
		TokenOptions: &auth.TokenOptions{
			Secret: "secret",
		},
	}

	serveCmd.Server.APIConfigurator = config.ConfigureAPI

	serveCmd.Server.SetLog(log)
	serveCmd.Server.ConfigureAPI()

	success = true

	return &APITester{
		t:       t,
		handler: serveCmd.Server.GetHandler(),
		api:     serveCmd.API,

		headers: make(map[string]string),

		TestLogger: testLog,
		Logger:     log,
		DB:         db,

		ctxcancel: cancel,

		Config: config,
	}
}

// sets the tester current time to another value.
// Warning: the API now does not change with this method.
// Only way of changing the API time is to recreate an APITester.
func (tester *APITester) SetNow(t time.Time) {
	tester.now = &t
}

// Resets the APITester current time.
func (tester *APITester) ResetNow() {
	tester.now = nil
}

// Gets the APITester current time.
func (tester *APITester) Now() time.Time {
	if tester.now != nil {
		return *tester.now
	}

	return time.Now()
}

// Debug() method checks the debug boolean.
// used to put Debug() on apitester calls, to see more informations.
func (tester *APITester) Debug() {
	tester.debug = true
}

// Close releases all underlying resources. MUST be called, preferably with defer.
func (tester *APITester) Close() {
	tester.api.PreServerShutdown()
	tester.api.ServerShutdown()
	tester.DB.Close()
	tester.ctxcancel()
}

func (tester *APITester) ServeHTTP(rw http.ResponseWriter, req *http.Request) {
	tester.handler.ServeHTTP(rw, req)
}

// HandleRequest pass a request through the handler.
func (tester *APITester) HandleRequest(req *http.Request) *httptest.ResponseRecorder {
	for key, value := range tester.headers {
		req.Header.Set(key, value)
	}

	tester.Logger.Debug().Msgf("headers: %v", req.Header)

	rr := httptest.NewRecorder()
	tester.ServeHTTP(rr, req)

	return rr
}

func (tester *APITester) Do(req *http.Request) (*http.Response, error) {
	var (
		reqBodyBuf    bytes.Buffer
		respHeaderBuf bytes.Buffer
	)

	if req.Body != nil {
		req.Body = TeeReadCloser(req.Body, &reqBodyBuf)
	}

	rr := tester.HandleRequest(req)

	// We disable bodyclose because it gives a false positive here
	response := rr.Result()
	defer response.Body.Close()

	respBody, err := io.ReadAll(response.Body)
	if err != nil {
		return nil, err
	}

	response.Body = io.NopCloser(bytes.NewBuffer(respBody))

	if err := response.Header.Write(&respHeaderBuf); err != nil {
		return nil, err
	}

	return response, nil
}

func (tester *APITester) APITest(name ...string) *apitest.APITest {
	var (
		reqBodyBuf    bytes.Buffer
		respHeaderBuf bytes.Buffer
	)

	return apitest.New(name...).
		Intercept(func(req *http.Request) {
			for key, value := range tester.headers {
				req.Header.Set(key, value)
			}
			if req.Body != nil {
				req.Body = TeeReadCloser(req.Body, &reqBodyBuf)
			}
		}).
		Observe(func(res *http.Response, req *http.Request, apiTest *apitest.APITest) {
			respBody, err := io.ReadAll(res.Body)
			if err != nil {
				tester.Logger.Err(err).Msg("could not read response body")
			}
			res.Body = io.NopCloser(bytes.NewBuffer(respBody))

			if err := res.Header.Write(&respHeaderBuf); err != nil {
				tester.Logger.Err(err).Msg("could not read response headers")
			}
		}).
		Handler(tester.handler)
}

// DoRequest creates and run request.
func (tester *APITester) DoRequest(verb, url, body string) (int, string, error) {
	return tester.DoRequestCtx(tester.t.Context(), verb, url, body)
}

// DoRequestCtx creates and run request.
// Needs a context.
func (tester *APITester) DoRequestCtx(ctx context.Context, verb, url, body string) (int, string, error) {
	var payload io.Reader
	if len(body) != 0 {
		payload = bytes.NewBufferString(body)
	}

	req, err := http.NewRequestWithContext(ctx, verb, url, payload)
	if err != nil {
		return 0, "", err
	}

	if payload != nil {
		req.Header.Set("Content-Type", "application/json")
	}

	// We disable bodyclose because it gives a false positive here
	response, _ := tester.Do(req)
	defer func() {
		assert.NoError(tester.t, response.Body.Close())
	}()

	b, err := io.ReadAll(response.Body)
	if err != nil {
		return 0, "", err
	}

	tester.Logger.Trace().Msg("Get response: " + string(b))

	return response.StatusCode, string(b), nil
}

// MustDoRequest ...
func (tester *APITester) MustDoRequest(verb, url, body string) (int, string) {
	code, body, err := tester.DoRequest(verb, url, body)
	if err != nil {
		tester.Logger.Err(err).Msg("")
		tester.t.Fatal(err)
	}

	return code, body
}

// DoJSONRequest creates and run request.
func (tester *APITester) DoJSONRequest(verb, url string, body interface{}, response interface{}) (int, error) {
	return tester.DoJSONRequestCtx(tester.t.Context(), verb, url, body, response)
}

// DoJSONRequestCtx creates and run request.
// Needs a context.
func (tester *APITester) DoJSONRequestCtx(
	ctx context.Context,
	verb, url string,
	body, response interface{},
) (int, error) {
	var sbody string

	if body != nil {
		if s, ok := body.(string); ok {
			sbody = s
		} else {
			b, err := json.Marshal(body)
			if err != nil {
				return 0, err
			}

			sbody = string(b)
		}
	}

	code, s, err := tester.DoRequestCtx(ctx, verb, url, sbody)
	if err != nil {
		return 0, err
	}

	if response != nil && len(s) != 0 {
		if err := json.Unmarshal([]byte(s), response); err != nil {
			// attempt to unmarshal an error
			var errorResponse models.Error
			if err := json.Unmarshal([]byte(s), &errorResponse); err == nil {
				return 0, errorResponse
			}

			return 0, err
		}
	}

	return code, nil
}

// MustDoJSONRequest creates and run request that takes no payload.
func (tester *APITester) MustDoJSONRequest(verb, url string, body interface{}, response interface{}) int {
	return tester.MustDoJSONRequestCtx(tester.t.Context(), verb, url, body, response)
}

// MustDoJSONRequestCtx creates and run request that takes no payload.
// Needs a context.
func (tester *APITester) MustDoJSONRequestCtx(
	ctx context.Context,
	verb, url string,
	body, response interface{},
) int {
	code, err := tester.DoJSONRequestCtx(ctx, verb, url, body, response)
	if err != nil {
		tester.Logger.Err(err).Msg("")
		tester.t.Fatal(err)
	}

	return code
}

// DoWithPayload creates and run request that takes a payload.
func (tester *APITester) DoWithPayload(verb, url string) {
}

// Get runs a GET request.
func (tester *APITester) Get(url string) (int, string, error) {
	return tester.DoRequest("GET", url, "")
}

// GetJSON runs a GET request and parse the result as JSON.
func (tester *APITester) GetJSON(url string, obj interface{}) (int, error) {
	return tester.DoJSONRequest("GET", url, nil, obj)
}

// MustGet runs a GET request an panics if it fails.
func (tester *APITester) MustGet(url string) (int, string) {
	return tester.MustDoRequest("GET", url, "")
}

// MustGetJSON runs a GET request, parse the result as JSON and panic on any error.
func (tester *APITester) MustGetJSON(url string, obj interface{}) int {
	return tester.MustGetJSONCtx(tester.t.Context(), url, obj)
}

// MustGetJSONCtx runs a GET request, parse the result as JSON and panic on any error.
// Needs a context.
func (tester *APITester) MustGetJSONCtx(ctx context.Context, url string, obj interface{}) int {
	return tester.MustDoJSONRequestCtx(ctx, "GET", url, nil, obj)
}

// Post runs a POST request.
func (tester *APITester) Post(url string, body string) (int, string, error) {
	return tester.DoRequest("POST", url, body)
}

// MustPost runs a POST request and panic if it fails.
func (tester *APITester) MustPost(url string, body string) (int, string) {
	return tester.MustDoRequest("POST", url, body)
}

// PostJSON runs a POST request with given input marshalled as JSON and responsed
// unmarshalled from JSON.
func (tester *APITester) PostJSON(url string, body interface{}, response interface{}) (int, error) {
	return tester.DoJSONRequest("POST", url, body, response)
}

// MustPostJSON is like PostJSON but stops the test on failure.
func (tester *APITester) MustPostJSON(
	url string,
	body, response interface{},
) int {
	return tester.MustPostJSONCtx(tester.t.Context(), url, body, response)
}

// MustPostJSONCtx is like PostJSON but stops the test on failure.
// Needs a context.
func (tester *APITester) MustPostJSONCtx(
	ctx context.Context,
	url string,
	body, response interface{},
) int {
	return tester.MustDoJSONRequestCtx(ctx, "POST", url, body, response)
}

// Put runs a POST request.
func (tester *APITester) Put(url string, body string) (int, string, error) {
	return tester.DoRequest("PUT", url, body)
}

// MustPut runs a POST request and panic if it fails.
func (tester *APITester) MustPut(url string, body string) (int, string) {
	return tester.MustDoRequest("PUT", url, body)
}

// PutJSON runs a POST request with given input marshalled as JSON and responsed
// unmarshalled from JSON.
func (tester *APITester) PutJSON(url string, body interface{}, response interface{}) (int, error) {
	return tester.DoJSONRequest("PUT", url, body, response)
}

// MustPutJSON is like PutJSON but stops the test on failure.
func (tester *APITester) MustPutJSON(url string, body interface{}, response interface{}) int {
	return tester.MustDoJSONRequest("PUT", url, body, response)
}

// SetHeader sets a header for all subsequent requests.
func (tester *APITester) SetHeader(key, value string) {
	if value != "" {
		tester.headers[key] = value
	} else {
		delete(tester.headers, key)
	}
}

// SetT switch the current t and returns a func that restore the initial one.
func (tester *APITester) SetT(t *testing.T) func() {
	t.Helper()

	save := tester.t
	tester.t = t
	restDB := tester.DB.SetTB(t)

	return func() {
		restDB()

		tester.t = save
	}
}

// Run runs the test as t.Run would, and call SetTB in addition.
func (tester *APITester) Run(name string, run func(t *testing.T)) bool {
	tester.t.Helper()

	return tester.t.Run(name, func(t *testing.T) {
		defer tester.TestLogger.SetTB(t)()
		defer tester.SetT(t)()
		run(t)
	})
}
